@@ -39,6 +39,14 @@
 #' See [internal-faq-ptype2-identity] for more information about
 #' identity values.
 #'
+#' `vec_ptype()` is a _performance_ generic. It is not necessary to implement it
+#' because the default method will work for any vctrs type. However the default
+#' method builds around other vctrs primitives like `vec_slice()` which incurs
+#' performance costs. If your class has a static prototype, you might consider
+#' implementing a custom `vec_ptype()` method that returns a constant. This will
+#' improve the performance of your class in many cases ([common
+#' type][vec_ptype2] imputation in particular).
+#'
 #' Because it may contain unspecified vectors, the prototype returned
 #' by `vec_ptype()` is said to be __unfinalised__. Call
 #' [vec_ptype_finalise()] to finalise it. Commonly you will need the
@@ -49,6 +57,13 @@
 #' successively calls [vec_ptype2()] to find a common type. It returns
 #' a [finalised][vec_ptype_finalise] prototype.
 #'
+#' @section Dependencies of `vec_ptype()`:
+#' - [vec_slice()] for returning an empty slice
+#'
+#' @section Dependencies of `vec_ptype_common()`:
+#' - [vec_ptype2()]
+#' - [vec_ptype_finalise()]
+#'
 #' @export
 #' @examples
 #' # Unknown types ------------------------------------------
@@ -87,7 +102,8 @@ vec_ptype <- function(x, ..., x_arg = "") {
   if (!missing(...)) {
     ellipsis::check_dots_empty()
   }
-  .Call(vctrs_ptype, x, x_arg)
+  return(.Call(vctrs_ptype, x, x_arg))
+  UseMethod("vec_ptype")
 }
 
 #' @export
@@ -95,10 +111,23 @@ vec_ptype <- function(x, ..., x_arg = "") {
 vec_ptype_common <- function(..., .ptype = NULL) {
   .External2(vctrs_type_common, .ptype)
 }
+vec_ptype_common_opts <- function(...,
+                                  .ptype = NULL,
+                                  .opts = fallback_opts()) {
+  .External2(vctrs_ptype_common_opts, .ptype, .opts)
+}
 vec_ptype_common_params <- function(...,
                                     .ptype = NULL,
-                                    .df_fallback = DF_FALLBACK_DEFAULT) {
-  .External2(vctrs_ptype_common_params, .ptype, .df_fallback)
+                                    .df_fallback = NULL,
+                                    .s3_fallback = NULL) {
+  opts <- fallback_opts(
+    df_fallback = .df_fallback,
+    s3_fallback = .s3_fallback
+  )
+  vec_ptype_common_opts(..., .ptype = .ptype, .opts = opts)
+}
+vec_ptype_common_fallback <- function(..., .ptype = NULL) {
+  vec_ptype_common_opts(..., .ptype = .ptype, .opts = full_fallback_opts())
 }
 
 #' @export
