@@ -343,12 +343,12 @@ test_that("as_unique_names() eliminates emptiness and duplication", {
   expect_identical(as_unique_names(x), c("...1", "x...2", "y", "x...4"))
 })
 
-test_that("solo empty or NA gets suffix", {
+test_that("as_unique_names(): solo empty or NA gets suffix", {
   expect_identical(as_unique_names(""), "...1")
   expect_identical(as_unique_names(NA_character_), "...1")
 })
 
-test_that("ellipsis treated like empty string", {
+test_that("as_unique_names() treats ellipsis like empty string", {
   expect_identical(as_unique_names("..."), as_unique_names(""))
 })
 
@@ -458,12 +458,12 @@ test_that("dupes get a suffix", {
   )
 })
 
-test_that("solo empty or NA gets suffix", {
+test_that("as_universal_names(): solo empty or NA gets suffix", {
   expect_identical(as_universal_names(""), "...1")
   expect_identical(as_universal_names(NA_character_), "...1")
 })
 
-test_that("ellipsis treated like empty string", {
+test_that("as_universal_names() treats ellipsis like empty string", {
   expect_identical(as_universal_names("..."), as_universal_names(""))
 })
 
@@ -778,6 +778,10 @@ test_that("NULL name specs works with scalars", {
   expect_identical(apply_name_spec(NULL, "foo", NULL, 1L), "foo")
   expect_named(vec_c(foo = 1), "foo")
 
+  expect_identical(apply_name_spec(NULL, "foo", chr(), 0L), chr())
+  expect_named(vec_c(foo = set_names(dbl())), chr())
+  expect_named(vec_c(foo = set_names(dbl()), bar = set_names(dbl())), chr())
+
   expect_error(apply_name_spec(NULL, "foo", c("a", "b")), "vector of length > 1")
   expect_error(vec_c(foo = c(a = 1, b = 2)), "vector of length > 1")
 
@@ -831,3 +835,20 @@ test_that("apply_name_spec() recycles return value not arguments (#1099)", {
   expect_identical(inner, c("a", "b", "c"))
   expect_identical(outer, "outer")
 })
+
+test_that("r_chr_paste_prefix() works", {
+  nms <- c("foo", "bar")
+
+  expect_equal(
+    .Call(vctrs_chr_paste_prefix, nms, "baz", "."),
+    c("baz.foo", "baz.bar")
+  )
+
+  # Greater than `VCTRS_PASTE_BUFFER_MAX_SIZE`
+  long_prefix <- strrep("a", 5000)
+
+  expect_equal(
+    .Call(vctrs_chr_paste_prefix, nms, long_prefix, "."),
+    paste0(long_prefix, ".", nms)
+  )
+})
